"""
These tests assume that twitter data has already been collected using tox -e
collect_test_data - the data is not stored in the repository to comply with
the Twitter terms of service and it's also necessary to use different
endpoints to capture the different shapes of API responses so rehydration
won't be helpful.

Note that the testing strategy here is close to end-to-end testing, exercising
the CLI as entrypoints for the functionality of the package.

"""
import pathlib
import sqlite3

from click.testing import CliRunner

from twittersphere import cli

data_path = pathlib.Path("tests", "data")


def test_prepare_filter_db(tmp_path):

    target_db = tmp_path / "test.db"

    runner = CliRunner()

    result = runner.invoke(
        cli.twittersphere,
        [
            "prepare",
            str(data_path / "test_sample.json"),
            str(data_path / "test_search_recent.json"),
            str(data_path / "test_hydrated_profiles.json"),
            str(target_db),
        ],
    )
    assert result.exit_code == 0

    db = sqlite3.connect(target_db)
    tweet_count = list(db.execute("select count(*) from directly_collected_tweet"))[0][
        0
    ]
    assert tweet_count >= 2000

    # Check that each of the referenced tweet fields are populated
    reply_count, quote_count, retweet_count = list(
        db.execute(
            "select count(replied_to_tweet_id), count(quoted_tweet_id), count(retweeted_tweet_id) from tweet_latest"
        )
    )[0]

    assert reply_count > 0 and quote_count > 0 and retweet_count > 0

    direct_user_count = list(
        db.execute("select count(*) from directly_collected_user")
    )[0][0]
    total_user_count = list(
        db.execute("select count(distinct user_id) from user_at_time")
    )[0][0]

    with open(data_path / "user_ids.txt", "r") as f:
        actual_user_count = len({l for l in f})

    # Note that users can be suspended in between collection and hydration!
    assert 0 < direct_user_count <= actual_user_count
    assert direct_user_count < total_user_count

    # Apply the simple "australia" rule to the database
    result = runner.invoke(
        cli.twittersphere,
        [
            "filter-users",
            str(data_path / "example_rule.csv"),
            str(target_db),
            "--ruleset-name",
            "test",
        ],
    )

    assert result.exit_code == 0
    matched_count = list(
        db.execute(
            """
            select count(*)
            from user_matching_ruleset
            where ruleset_name = 'test'
            """
        )
    )[0][0]

    assert matched_count > 0


# Test applying the rules to the database, and to the streams of tweets
# Test extracting the qa profiles?
# Test applying the labelled qa profiles.



import json

with open("tests/data/test_sample.json", "rb") as f:
    with open("tests/data/user_ids.txt", "w") as out:
        for line in f:
            author_id = json.loads(line)["data"]["author_id"]
            out.write(author_id)
            out.write("\n")



"""
The core of the twittersphere package for ingesting tweets and creating or
applying content based rulesets.

"""
from collections import defaultdict, Counter
import csv
import json
import random
import sys

import click
import regex


token_regex = regex.compile(
    r"\b\p{Word_Break=WSegSpace}*['\u200D]?",
    flags=regex.WORD | regex.UNICODE | regex.V1,
)

stopwords = set(
    ", . ' \" - + & / | : ; \r\n \n \t \r ( ) ! * ? ] [ ^ ~ $ % { } =".split(" ")
)
stopwords.add("")


def tokenize(text, stopwords):
    """Return a list of tokens in text, after removing common punctuation."""
    return [t for t in token_regex.split(text.lower()) if t not in stopwords]


def ngrams(text, n, stopwords):
    """
    Return all ngrams up to size n in the string.

    Stopwords are removed from the stream of tokens before ngrams are generated.

    """
    token_list = tokenize(text, stopwords)
    n_tokens = len(token_list)
    found_ngrams = []

    for l in range(1, n + 1):

        offset = l - 1

        for i in range(0, n_tokens - offset):
            ngram = token_list[i : i + l]
            found_ngrams.append(tuple(ngram))

    return set(found_ngrams)


def deduplicate_rules(rules):
    """
    Given a set of rules that consist of ngrams, reduce them down to a minimal set.

    If the single token ('Australia',) is a rule, then even though
    ('Brisbane', 'Australia') is also a match, it is redundant. This procedures removes
    redundant rules of length n + 1 ththatat are already matched by rules of length n.

    Deduplication is necessary so we can properly account for multiple includes and
    excludes in the same tweet.
    """

    why_redundant = defaultdict(set)

    # Yeah, if you want to do longer length rules this should be properly broken down...
    for rule in rules:
        if len(rule) == 2:
            sub_rules = [tuple(rule)]
            # Check unigrams
            if rule[:1] in rules:
                why_redundant[rule[:1]].add(rule)
            if rule[1:] in rules:
                why_redundant[rule[1:]].add(rule)

        elif len(rule) == 3:
            # Check unigrams
            if rule[:1] in rules:
                why_redundant[rule[:1]].add(rule)
            if rule[1:2] in rules:
                why_redundant[rule[1:2]].add(rule)
            if rule[2:] in rules:
                why_redundant[rule[2:]].add(rule)

            # Check bigrams
            if rule[:2] in rules:
                why_redundant[rule[:2]].add(rule)
            if rule[1:] in rules:
                why_redundant[rule[1:]].add(rule)

    redundant_rules = set(rule for leaves in why_redundant.values() for rule in leaves)

    return rules - redundant_rules, why_redundant


def load_rule_files(file_paths):
    """
    Load a set of rules from the given files.

    A rule file is a CSV file containing at least the following fields
    in any order, with the following values:

    - include: 1, -1 or blank.

        1 includes that this is a positive matching rule, -1 a negative
        matching rule, blank for not a match. Blank fields won't affect
        the results, but are retained in the file to avoiding annotating
        rules again.

    - field: 'location', 'description' or 'realname'

        The field to match on.

    - first_token, second_token, third_token: strings

        tokens to match on as part of the rule. Single/double
        word matches should start on first_token and leave unneeded
        tokens blank.

    """
    redundant_include_rules = defaultdict(set)
    redundant_exclude_rules = defaultdict(set)

    # Used to not present rows that have already been presented to users for labelling,
    # and for ensuring that redundant labels are carried through always.
    all_rules = defaultdict(dict)

    for file in file_paths:
        with open(file, "r") as f:
            reader = csv.DictReader(f)

            fields = ["first_token", "second_token", "third_token"]

            for row in reader:

                ngram = tuple([row[f] for f in fields if row[f]])

                include = row["include"]
                field = row["field"]

                if include == "1":
                    redundant_include_rules[field].add(ngram)

                elif include == "-1":
                    redundant_exclude_rules[field].add(ngram)

                elif include:
                    raise ValueError(f"Unknown include column value {include}")

                all_rules[field][ngram] = include

                # Ensure the field exists in every defaultdict for downstream
                redundant_include_rules[field]
                redundant_exclude_rules[field]

    # Prune any exclude rules that don't have a subcomponent that doesn't match an
    # include rule. Note this needs to be done before deduplication.
    prune_rules = defaultdict(set)

    for field, rules in redundant_exclude_rules.items():
        for rule in rules:
            # By definition there's no associated positive rule
            if len(rule) == 1:
                prune_rules[field].add(rule)
            else:
                matched = False
                for token in rule:
                    if (token,) in redundant_include_rules[field]:
                        matched = True
                        break

                if len(rule) == 3 and (
                    rule[:2] in redundant_include_rules[field]
                    or rule[1:] in redundant_include_rules[field]
                ):
                    matched = True

                if not matched:
                    prune_rules[field].add(rule)

    for field, rules in prune_rules.items():
        redundant_exclude_rules[field] -= rules

    include_rules = {
        field: deduplicate_rules(rules)[0]
        for field, rules in redundant_include_rules.items()
    }

    exclude_rules = {
        field: deduplicate_rules(rules)[0]
        for field, rules in redundant_exclude_rules.items()
    }

    return include_rules, exclude_rules, all_rules


def classify_user_profile(profile, include_rules, exclude_rules):
    """
    Apply the rules to the given profile.

    Returns the boolean match, and which rules matched, although the
    later is not currently used.
    """

    positive_matches = defaultdict(set)
    negative_matches = defaultdict(set)

    for field in ("name", "description", "location"):

        text = profile.get(field, "")

        ngram_tokens = ngrams(text, 3, stopwords=stopwords)
        positive_matches[field] = ngram_tokens & include_rules.get(field, set())
        negative_matches[field] = ngram_tokens & exclude_rules.get(field, set())

    matched = sum(len(m) for m in positive_matches.values()) > sum(
        len(m) for m in negative_matches.values()
    )

    return matched, positive_matches, negative_matches


def apply_rules(api_result_page, include_rules, exclude_rules):
    """
    Take a set of rules, and apply it to a page of V2 Twitter API data.

    Returns a new JSON document, containing the subset of tweets (or accounts)
    directly authored by a profile considered to match the given rules.

    Also returns data structures that allow accumulating counts of tweets by
    user_id for the purpose of understanding distributions and quantifying
    effectiveness.

    """

    matched_author_ids = set()
    transformed_result_page = api_result_page.copy()

    # Note the granularity here is author_id -> profile fields.
    # There is only one user object per page in the API responses,
    # but this is not the right level for the total evaluation once
    # results from pages are merged.
    account_tweet_count = dict()

    # Tweets have authors, user profile objects don't
    if "author_id" in api_result_page["data"][0]:

        # Authored tweets - we need to extract this here, as we're currently only labelling
        # top level collected tweets, not tweets included by reference.
        top_level_authors = set(tweet["author_id"] for tweet in api_result_page["data"])

        for profile in api_result_page["includes"]["users"]:

            author_id = profile["id"]

            # Skip user objects for authors of referenced tweets.
            if author_id not in top_level_authors:
                continue

            matched, positive_matches, negative_matches = classify_user_profile(
                profile, include_rules, exclude_rules
            )

            author_fields = tuple(
                profile.get(field, "") for field in ("name", "description", "location")
            )

            if matched:
                matched_author_ids.add(author_id)
                account_tweet_count[(True, author_id)] = [author_fields, 0]
            else:
                account_tweet_count[(False, author_id)] = [author_fields, 0]

        transformed_result_page["data"] = []

        for tweet in api_result_page["data"]:

            author_id = tweet["author_id"]

            if author_id in matched_author_ids:

                transformed_result_page["data"].append(tweet)
                account_tweet_count[(True, author_id)][1] += 1

            else:
                account_tweet_count[(False, author_id)][1] += 1

    # User profiles
    else:
        raise ValueError("Only tweets are currently supported for filtering")

    return transformed_result_page, account_tweet_count


def transform_profiles_for_annotation(
    account_tweet_count, qa_profile_count, sampling="class_stratified"
):
    if qa_profile_count < 0:
        raise ValueError(
            f"qa_profile_count should be positive, instead it was {qa_profile_count}."
        )

    if sampling == "class_stratified":
        matched_profile_lines = [
            key for key, (_, _, matched) in account_tweet_count.items() if matched
        ]
        not_matched_profile_lines = [
            key for key, (_, _, matched) in account_tweet_count.items() if not matched
        ]

        sample_profiles = random.sample(
            matched_profile_lines, qa_profile_count // 2
        ) + random.sample(not_matched_profile_lines, qa_profile_count // 2)

    else:
        sample_profiles = random.sample(
            list(account_tweet_count.keys()), qa_profile_count
        )

    random.shuffle(sample_profiles)

    bit_location = 2**16

    for profile in sample_profiles:
        author_ids, tweet_count, matched = account_tweet_count[profile]

        # We're going to obfuscate the matched column with a random number so
        # it doesn't leak into labelling
        obfuscated_match = random.randint(0, sys.maxsize)
        if matched:
            obfuscated_match |= bit_location
        else:
            obfuscated_match & ~bit_location

        yield (*profile, len(author_ids), tweet_count, hex(obfuscated_match))


def apply_user_rules_to_db(db_conn, include_rules, exclude_rules, ruleset_name):
    """
    Apply the specified include/exclude rules to the given database conn.

    Only the latest version of each profile is considered for labelling.

    """
    try:
        db_conn.execute("begin")

        db_conn.execute(
            "delete from user_matching_ruleset where ruleset_name = ?", [ruleset_name]
        )

        profiles = db_conn.execute(
            "select user_id, location, description, name from user_latest"
        )

        for user_id, location, description, name in profiles:
            matched, _, _ = classify_user_profile(
                {
                    "location": location or "",
                    "description": description or "",
                    "name": name or "",
                },
                include_rules,
                exclude_rules,
            )

            if matched:
                db_conn.execute(
                    "insert into user_matching_ruleset values (?, ?)",
                    [ruleset_name, user_id],
                )

        db_conn.execute("commit")

    except:
        db_conn.execute("rollback")
        raise


def count_user_ngrams(db_conn, stopwords, ruleset_name=""):
    """
    Count ngrams present in the latest version of each user in the collection.

    If no ruleset_name is provided, this will be a global count across all
    matching users, otherwise only the `user_id`'s that match `ruleset_name`
    in the `user_matching_ruleset` table will be counted. This is to support
    statistical comparisons across different ruleset groups without
    tokenising everything again.

    """

    try:
        db_conn.execute("begin")

        db_conn.execute(
            """
            create temporary table ruleset_ngram_part as
                select *
                from user_ruleset_ngram_count
                limit 0
            """
        )

        if ruleset_name:
            profiles = db_conn.execute(
                """
                select
                    user_id, location, description, name
                from user_latest
                where user_id in (
                    select user_id
                    from user_matching_ruleset
                    where ruleset_name = ?
                )
                """,
                [ruleset_name],
            )
        else:
            profiles = db_conn.execute(
                """
                select
                    user_id, location, description, name
                from user_latest
                """
            )

        ngram_counts = defaultdict(Counter)

        counts = 0

        for user_id, location, description, name in profiles:

            # Tokenise all of the fields.
            if location:

                l = ngrams(location, 3, stopwords=stopwords)
                counts += len(l)

                for ngram in l:
                    ngram_counts["location"][ngram] += 1

            if description:

                d = ngrams(description, 3, stopwords=stopwords)
                counts += len(d)

                for ngram in d:
                    ngram_counts["description"][ngram] += 1

            if name:

                n = ngrams(name, 3, stopwords=stopwords)
                counts += len(n)

                for ngram in n:
                    ngram_counts["name"][ngram] += 1

            if counts >= 1000000:

                db_conn.executemany(
                    """insert into ruleset_ngram_part values(?, ?, ?, ?, ?, ?)""",
                    (
                        (ruleset_name, field, *(n + (("",) * (3 - len(n)))), count)
                        for field, values in ngram_counts.items()
                        for n, count in values.items()
                    ),
                )

                ngram_counts = defaultdict(Counter)
                counts = 0

        # Make sure to get that final batch.
        db_conn.executemany(
            """insert into ruleset_ngram_part values(?, ?, ?, ?, ?, ?)""",
            (
                (ruleset_name, field, *(n + (("",) * (3 - len(n)))), count)
                for field, values in ngram_counts.items()
                for n, count in values.items()
            ),
        )

        # Make sure to clear out old ngrams first
        db_conn.execute(
            "delete from user_ruleset_ngram_count where ruleset_name = ?",
            [ruleset_name],
        )

        # Aggregate the new parts into the table.
        db_conn.execute(
            """
            insert into user_ruleset_ngram_count
                select
                    ruleset_name,
                    field,
                    first_token,
                    second_token,
                    third_token,
                    sum(profile_count)
                from ruleset_ngram_part
                group by 1, 2, 3, 4, 5
            """
        )

        db_conn.execute("drop table ruleset_ngram_part")
        db_conn.execute("commit")

    except:
        db_conn.execute("rollback")
        db_conn.execute("drop table ruleset_ngram_part")
        raise



"""
This module is responsible for taking and storing Twitter related entities as
extracted in process.py into an SQLite database.

"""

import concurrent.futures as cf
import sqlite3
import threading
import time

from . import process, _schema, twittersphere


class SchemaVersionError(Exception):
    pass


def ensure_db(db_path):
    """
    Return an SQLite connection with appropriate schema at the given path.

    If the schema version ondisk is out of date an error will be raised -
    currently there is no option to migrate forward, data must be
    reinserted.

    """
    conn = sqlite3.connect(db_path, isolation_level=None, check_same_thread=False)

    try:
        schema_version = list(
            conn.execute(
                """
                SELECT value
                from metadata
                where key = 'twittersphere_schema_version'
                """
            )
        )[0][0]
    except sqlite3.OperationalError:
        schema_version = 0

    if schema_version == 0:
        # Execute migrations
        conn.execute("pragma journal_mode=WAL")
        conn.execute("begin")
        for statement in _schema.SCHEMA_STATEMENTS.split(";\n\n"):
            conn.execute(statement)

        conn.execute("commit")

    elif schema_version != _schema.CURRENT_SCHEMA_VERSION:
        raise SchemaVersionError(
            f"The on disk database has schema version {schema_version}, the "
            f"current version is {_schema.CURRENT_SCHEMA_VERSION}. You will "
            "need to create a new database and reinsert your data to use this "
            "version of twittersphere."
        )

    return conn


def flush_db(db_conn, target_db_path):
    """
    Flush the contents of db_conn into target_db_path, using SQLite 'attach'
    and insert in sorted primary key order.

    This is used to optimise insert using staging to an in memory database and
    periodic flushing to disk.

    """

    db_conn.execute("attach ? as flush_to", (target_db_path,))
    db_conn.execute("begin")

    for table, primary_key in _schema.table_keys.items():
        db_conn.execute(
            f"""
            insert or ignore into flush_to.{table}
            select *
            from main.{table}
            order by {primary_key}
            """
        )
    db_conn.execute("commit")
    db_conn.close()


def insert_pages(db_path, raw_pages, n_cpus=2, in_memory_max_db_size=2**31):
    """
    Process the stream of pages and insert them into the database.

    Data is staged to an in memory database until it hits
    `in_memory_max_db_size`, then it is flushed to the target database. The
    default size is 2GiB. For larger collections, using more memory is likely
    to make the process go faster.

    """
    futures = set()

    # Ensure the path exists for inserting with the correct schema.
    check_db = ensure_db(db_path)
    check_db.close()

    db_conn = ensure_db(":memory:")

    target = None
    batch_size = 0
    batch = []

    # Dispatch batches of 1 megabyte of raw content at a time.
    min_batch_size = 2**20

    with cf.ProcessPoolExecutor(n_cpus) as pool:

        db_conn.execute("begin")

        for raw_page in raw_pages:

            batch.append(raw_page)
            batch_size += len(raw_page)

            if batch_size >= min_batch_size:
                futures.add(pool.submit(process.process_pages, batch))
                batch = []
                batch_size = 0

            if len(futures) > n_cpus + 1:
                done, futures = cf.wait(futures, return_when="FIRST_COMPLETED")

                for f in done:
                    results_to_insert = f.result()
                    for result in results_to_insert:
                        insert_processed(db_conn, result)

                # Commit periodically from the in memory database
                db_size = list(
                    db_conn.execute(
                        """
                        select
                            (select page_count from pragma_page_count) *
                            (select page_size from pragma_page_size)
                        """
                    )
                )[0][0]

                if db_size >= in_memory_max_db_size:
                    db_conn.execute("commit")
                    # Wait for the previous db flush to finish
                    if target is not None:
                        target.join()

                    flush_db_conn = db_conn
                    target = threading.Thread(
                        target=flush_db, args=(flush_db_conn, db_path)
                    )
                    target.start()

                    # Start a new in memory database
                    db_conn = ensure_db(":memory:")
                    db_conn.execute("begin")

        futures.add(pool.submit(process.process_pages, batch))

        for f in cf.as_completed(futures):
            results_to_insert = f.result()
            for result in results_to_insert:
                insert_processed(db_conn, result)

        db_conn.execute("commit")

        if target is not None:
            target.join()

        flush_db(db_conn, db_path)


def insert_processed(db_conn, processed):
    """
    Insert processed data into the database.

    A savepoint is wrapped around this transaction to ensure that partial
    writes aren't seen.

    """
    data = processed["data"]
    includes = processed["includes"]
    metadata = processed["metadata"]

    # First, create the collection context id/metadata
    db_conn.execute(
        """
        insert or ignore into collection_context(
            retrieved_at,
            twitter_url,
            twarc_version
        ) values(:retrieved_at, :twitter_url, :twarc_version)
        """,
        metadata,
    )

    context_id = list(
        db_conn.execute(
            """
            select context_id
            from collection_context
            where (retrieved_at, twitter_url, twarc_version) =
                (:retrieved_at, :twitter_url, :twarc_version)
            """,
            metadata,
        )
    )[0][0]
    metadata["context_id"] = context_id

    # Process users
    users = data.get("users", []) + includes["users"]
    db_conn.executemany(
        """
        insert or ignore into user_at_time values (
            :id,
            :context_id,
            :retrieved_at,
            :name,
            :profile_image_url,
            :created_at,
            :protected,
            :description,
            :location,
            :pinned_tweet_id,
            :verified,
            :url,
            :username,
            :followers_count,
            :following_count,
            :tweet_count,
            :listed_count,
            :withheld_country_codes
        )
        """,
        (user | metadata for user in users),
    )

    db_conn.executemany(
        "insert or ignore into directly_collected_user values(:id)",
        (data.get("users", [])),
    )

    # Process Tweets
    tweets = data.get("tweets", []) + includes["tweets"]
    db_conn.executemany(
        """
        insert or ignore into tweet_at_time values (
            :id,
            :context_id,
            :author_id,
            :created_at,
            :retrieved_at,
            :conversation_id,
            :retweeted,
            :quoted,
            :replied,
            :text,
            :lang,
            :source,
            :possibly_sensitive,
            :reply_settings,
            :like_count,
            :quote_count,
            :reply_count,
            :retweet_count,
            :withheld_copyright,
            :withheld_country_codes,
            :poll_id,
            :place_id
        )
        """,
        (tweet | metadata for tweet in tweets),
    )

    # Tweet ancillary - mentions, media, urls, hashtags, annotations
    db_conn.executemany(
        "insert or ignore into tweet_media values (?, ?, ?)",
        (
            (tweet["id"], metadata["retrieved_at"], media_key)
            for tweet in tweets
            for media_key in tweet["media_keys"]
        ),
    )

    db_conn.executemany(
        "insert or ignore into tweet_hashtag values (?, ?, ?)",
        (
            (tweet["id"], metadata["retrieved_at"], hashtag)
            for tweet in tweets
            for hashtag in tweet["hashtags"]
        ),
    )

    db_conn.executemany(
        "insert or ignore into tweet_cashtag values (?, ?, ?)",
        (
            (tweet["id"], metadata["retrieved_at"], cashtag)
            for tweet in tweets
            for cashtag in tweet["cashtags"]
        ),
    )

    db_conn.executemany(
        "insert or ignore into tweet_mention values (?, ?, ?, ?)",
        (
            (
                tweet["id"],
                metadata["retrieved_at"],
                mention["user_id"],
                mention["username"],
            )
            for tweet in tweets
            for mention in tweet["mentions"]
        ),
    )

    db_conn.executemany(
        """
        insert or ignore into url values (
            :url,
            :retrieved_at,
            :description,
            :display_url,
            :expanded_url,
            :images,
            :media_key,
            :status,
            :title,
            :unwound_url
        )
        """,
        (url | metadata for tweet in tweets for url in tweet["urls"]),
    )

    db_conn.executemany(
        """
        insert or ignore into tweet_url values (
            :id,
            :retrieved_at,
            :url
        )
        """,
        (
            (url["url"], metadata["retrieved_at"], tweet["id"])
            for tweet in tweets
            for url in tweet["urls"]
        ),
    )

    db_conn.executemany(
        "insert or ignore into directly_collected_tweet values(:id)",
        (data.get("tweets", [])),
    )

    db_conn.executemany(
        "insert or ignore into tweet_entity_domain values (?, ?, ?, ?)",
        (
            (
                tweet["id"],
                metadata["retrieved_at"],
                annotation["entity_id"],
                annotation["domain_id"],
            )
            for tweet in tweets
            for annotation in tweet["context_annotations"]
        ),
    )

    db_conn.executemany(
        "insert or ignore into entity values (:entity_id, :entity_name, :entity_description)",
        (annotation for tweet in tweets for annotation in tweet["context_annotations"]),
    )

    db_conn.executemany(
        "insert or ignore into domain values (:domain_id, :domain_name, :domain_description)",
        (annotation for tweet in tweets for annotation in tweet["context_annotations"]),
    )

    # Process polls
    db_conn.executemany(
        """
        insert or ignore into poll values (
            :id,
            :retrieved_at,
            :duration_minutes,
            :end_datetime,
            :voting_status
        )
        """,
        (poll | metadata for poll in includes["polls"]),
    )

    db_conn.executemany(
        """
        insert or ignore into poll_option values (
            :id,
            :retrieved_at,
            :position,
            :label,
            :votes
        )
        """,
        (
            option | poll | metadata
            for poll in includes["polls"]
            for option in poll["options"]
        ),
    )

    # Process Places
    db_conn.executemany(
        """
        insert or ignore into place values (
            :id,
            :country,
            :country_code,
            :full_name,
            :geo_type,
            :geo_bbox_1,
            :geo_bbox_2,
            :geo_bbox_3,
            :geo_bbox_4,
            :name,
            :place_type
        )
        """,
        includes["places"],
    )

    # Process media
    db_conn.executemany(
        """
        insert or ignore into media values (
            :media_key,
            :retrieved_at,
            :alt_text,
            :duration_ms,
            :preview_image_url,
            :view_count,
            :type,
            :url,
            :width,
            :height
        )
        """,
        (media | metadata for media in includes["media"]),
    )



import setuptools

install_requires = ["click>=8.0.0", "regex>=2021.3.17", "glom>=22.1.0"]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="twittersphere",
    author="Sam Hames and QUT Digital Observatory",
    author_email="s.hames@uq.edu.au",
    description="Tools for content based filtering of tweets and Twitter accounts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/QUT-Digital-Observatory/twittersphere",
    license="MIT",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Sociology",
    ],
    keywords="social_science social_media_analysis",
    python_requires=">=3.9",
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    install_requires=install_requires,
    entry_points={"console_scripts": ["twittersphere=twittersphere:cli.twittersphere"]},
)
